import random
import copy
from pyalgo import basic_modules
from pyalgo.basic_modules import sortalgo_default_values


def in_order(List, order='SmallToBig'):
    """
    this function takes a list and check if all the elements are in order
    the default is from the smallest to the biggest
    """

    # // check if the input is valid \\
    if order not in basic_modules.sortalgo_default_values.in_order_default_values:
        raise ValueError(
            f"'order' needs to be from the list of options: {sortalgo_default_values.in_order_default_values}")
        #raise ValueError("'order' needs to be from the list of possibilities:",
        #                 basic_modules.default_functions.print_list_arguments(sortalgo_default_values.in_order_default_values, print_the_msg=True, return_msg=True))

    ListType = basic_modules.default_values.ListType

    if not isinstance(List, ListType):
        raise TypeError("'List' needs to be a list type")



    length = len(List)  # get the length of the list (the number of check that we need to do)

    if order == 'SmallToBig':
        for i in range(length - 1):
            if (List[i + 1] < List[i]):
                return False
    elif order == 'BigToSmall':
        for i in range(length - 1):
            if (List[i + 1] > List[i]):
                return False

    return True





def quick_sort(List):   # // needs to be optimized \\
                        # // may  use a regeession \\

    def split_to_subsets(List, ConstNumber, move_same_value_to='down'):

        # // check that the input is valid \\

        if not move_same_value_to in basic_modules.sortalgo_default_values.\
                quick_sort_default_values.move_same_value_to_default_values:
            raise ValueError("'move_same_value_to' needs to 'up' or 'down'")


        List = copy.deepcopy(List)                 # make a deep-copy to not change the parameters outside the function
        ConstNumber = copy.deepcopy(ConstNumber)   # make a deep-copy to not change the parameters outside the function



        transferList = [[], [ConstNumber], []]

        # // Spliting between the 3 subsets \\
        # // subset 1: the numbers that are smaller than the choosen number \\
        # // subset 2: the choosen number itself \\
        # // subset 3: the numbers that are bigger than the choosen number \\
        for i in range(len(List)):
            if List[i] > ConstNumber:
                transferList[2].append(List[i])
            elif List[i] < ConstNumber:
                transferList[0].append(List[i])
            elif List[i] == ConstNumber:
                if move_same_value_to == 'down':
                    transferList[0].append(List[i])
                elif move_same_value_to == 'up':
                    transferList[2].append(List[i])

        # // Clearing the empty sublists \\
        for i in range(3):
            try:
                if transferList[i] == []:
                    del transferList[i]
            except: pass

        return transferList

    # // preparing list \\
    List = basic_modules.default_functions.to_single_list(List)

    Random_Choosen_Number = random.randint(0, len(List) - 1)
    Random_Choosen_Number = List[Random_Choosen_Number]

    List.remove(Random_Choosen_Number)

    List = split_to_subsets(List, Random_Choosen_Number)

    while not in_order(basic_modules.default_functions.to_single_list(List)):
        for i in range(len(List)):
            if not in_order(basic_modules.default_functions.to_single_list(List[i])):
                pass
                Random_Choosen_Number = random.randint(0, len(List[i])-1)
                Random_Choosen_Number = List[i][Random_Choosen_Number]

                List[i].remove(Random_Choosen_Number)

                TransferList = split_to_subsets(List[i], Random_Choosen_Number)

                del List[i]
                for j in range(len(TransferList)):
                    List.insert(i, TransferList[-1-j])

    return basic_modules.default_functions.to_single_list(List)


from . import default_values
import copy
from numba import njit



def to_single_list(List):  # // needs to be optimized \\
                           # // may use a regression \\

    """
    This function takes a list and moving out the
    sub lists until there is no sublists
    """

    List = copy.deepcopy(List)  # deep copy the list to not make a reference

    dumbList = []
    ListType = type(dumbList)

    if not isinstance(List, ListType): raise TypeError("You need to enter a list")

    is_single_list = True

    while is_single_list:
        for i in range(len(List)):
            if isinstance(List[i], ListType):

                is_single_list = False

                length = len(List[i])
                transferList = List[i]

                del List[i]

                for j in range(length):
                    List.insert(i, transferList[-1-j])

        is_single_list = not is_single_list

    return List


def print_list_arguments(List, Tag="-->", print_the_msg=True, return_msg=True):

    """
    This function takes a list and prints its arguments separated with the Tag first

    example:

    option number {the number of the argument}:  {Tag} {The arguments}

    :param List: this should be a list
    :param Tag: This should be a string
    :return: this function doesn't returns anything
    """

    msg = ""

    # // check that the type of the inputs are valid to this function \\
    if not isinstance(List, default_values.ListType):
        raise TypeError("'List' should be a list type")
    if not isinstance(Tag, default_values.StringType):
        raise TypeError("'Tag' should be a String type")


    for i in range(len(List)):
        if print_the_msg:
            print("option number {0}:  {1} {2}".format(i+1, Tag, List[i]))

        msg += "option number {0}:  {1} {2}".format(i+1, Tag, List[i])
        msg += "\n"

    if return_msg: return msg



def gcd(a, b):

    # this functions uses recursion to fing the gcd of a, b

    """
    This function returns the greatest common divisor
    """

    if a < 0:  raise ValueError("'a' should be greater than 0")
    if b < 0:  raise ValueError("'b' should be greater than 0")

    if b > a:
        a, b = b, a

    if b == 0:
        return a

    return gcd(b, a % b)

@njit
def prime_check(n):
    """Primality test using 6k+-1 optimization."""

    if n <= 3:
        return n > 1

    if ((n%2 == 0) or (n%3 == 0)):
        return False

    i = 5
    while (i ** 2) <= n:
        if ((n%i == 0) or (n%(i+2) == 0)):
            return False

        i += 6
    return True


"""
This function is contains the default
variables for the 'sortalgo' module
"""
in_order_default_values = ['SmallToBig', 'BigToSmall']

#class quick_sort_default_values:
#    def __init__(self): pass
#
#    move_same_value_to_default_values = ['down', 'up']


import numpy as np

BoolType = type(True)

StringType = type("This is a string")
CharType = type('s')

IntType = type(5)
FloatType = type(5.5)

ListType = type([])
DictionatyType = type({1:"first", "2":"second", 3:"third"})

ByteType = type(b"hello there")
ByteArrayType = type(bytearray(b"hello there"))

ArrayType = type(np.array([]))


print("...DONE...")


# python -m pip install --user --upgrade setuptools wheel
# python -m pip install --user --upgrade twine

# python setup.py sdist bdist_wheel

# python -m twine upload --repository testpypi dist/*  <-- upload to test Pypi
# python -m twine upload dist/*  <-- upload to Pypi




#You can use
#[Github-flavored Markdown](https://guides.github.com/features/mastering-markdown/)
#to write your content.





# // Mission Log \\
#
# pytrainer --> ML_functions // add lambda to Normal_Equation
# pytrainer --> ML_functions // complete Cost_function_Grad -> linear regression \\
# pytrainer --> ML_functions // complete Cost_function -> logistic regression \\
# pytrainer --> ML_functions // complete Cost_function_Grad -> logistic regression \\
#
# pytrainer --> ML_functions // complete gradient_descent \\
# pytrainer --> ML_functions_Training // complete forward_propagation \\ (add forward propagation to neural network)


from pyalgo.basic_modules import default_functions, default_values
from . import compiled_functions


class XOR_Cipher:
    def __init__(self, key, description: str = ""):
        self.check_if_key_is_valid(key)

        self.key = key

        self.description = description

    def encrypt_bytes(self, bytes, key=None):
        ByteType = default_values.ByteType
        ByteArrayType = default_values.ByteArrayType

        if not (isinstance(bytes, ByteType) or isinstance(bytes, ByteArrayType)):
            raise TypeError("'bytes' needs to be byte type or bytearray type")

        if key == None:
            key = self.key
        else:
            self.check_if_key_is_valid(key)

        if isinstance(bytes, ByteType):
            bytes = bytearray(bytes)

        for i, value in enumerate(bytes):
            if isinstance(key, default_values.IntType):
                bytes[i] = key ^ value
            else:
                bytes[i] = key[i % len(key)] ^ value

        return bytes

    def decrypt_bytes(self, bytes, key=None):
        ByteType = default_values.ByteType
        ByteArrayType = default_values.ByteArrayType

        if not (isinstance(bytes, ByteType) or isinstance(bytes, ByteArrayType)):
            raise TypeError("'bytes' needs to be byte type or bytearray type")

        if key == None:
            key = self.key
        else:
            self.check_if_key_is_valid(key)

        if isinstance(bytes, ByteType):
            bytes = bytearray(bytes)

        for i, value in enumerate(bytes):
            if isinstance(key, default_values.IntType):
                bytes[i] = key ^ value
            else:
                bytes[i] = key[i % len(key)] ^ value

        return bytes

    def encrypt_file(self, path, key: int = None, include_additional_end: bool = True, additional_end="enc"):
        if key == None:
            key = self.key
        else:
            self.check_if_key_is_valid(key)

        with open(path, "br") as file:
            bytes = file.read()

        bytes = self.encrypt_bytes(bytes, key)

        if include_additional_end:
            path += additional_end

        with open(path, "bw") as file:
            file.write(bytes)

        print("the file has encrypted properly")

    def decrypt_file(self, path, key=None, remove_additional_end: bool = True, additional_end="enc"):
        if key == None:
            key = self.key
        else:
            self.check_if_key_is_valid(key)

        with open(path, "br") as file:
            bytes = file.read()

        bytes = self.decrypt_bytes(bytes, key)

        if remove_additional_end:
            path = path[:-len(additional_end)]

        with open(path, "bw") as file:
            file.write(bytes)

        print("the file has decrypted properly")

    @staticmethod
    def check_if_key_is_valid(key):
        IntType, ListType = default_values.IntType, default_values.ListType

        if not (isinstance(key, IntType) or isinstance(key, ListType)):
            raise TypeError("'key' can be an int type or a list type")

        if isinstance(key, IntType):
            if key >= 256:
                raise ValueError("'key' cannot be bigger than 255")

            if key < 0:
                raise ValueError("'key' cannot be smaller than 0")
        elif isinstance(key, ListType):
            for i in key:
                if i >= 256:
                    raise ValueError("any number in 'key' cannot be bigger than 255")

                if i < 0:
                    raise ValueError("any number in 'key' cannot be smaller than 0")


class RSA():
    def __init__(self, p, q, e, description: str):
        self.p = p
        self.q = q
        self.e = e

        self.Phi = ((self.p - 1) * (self.q - 1))

        self.description = description

        if e >= self.Phi:  raise ValueError(f"'e' cannot be bigger that phi({self.Phi})")

        self.check_if_e_is_valid(e, p, q)
        if default_functions.prime_check(e) == False:  raise ValueError("'e' should be prime")
        if default_functions.prime_check(p) == False:  raise ValueError("'p' should be prime")
        if default_functions.prime_check(q) == False:  raise ValueError("'q' should be prime")

        self.N = p * q  # compute N

    def check_if_e_is_valid(self, e=None, p=None, q=None):

        if e == None:  e == self.e
        if p == None:  p == self.p
        if q == None:  q == self.q
        if not (default_functions.gcd(e, self.Phi) == 1):
            raise ValueError("'e' should not have a gcd with φ(p, q) that is not 1")

    def find_private_key(self):
        self.d = compiled_functions.find_private_key(self.e, self.Phi)

    def encrypt(self, M, N=None, e=None):

        StringType = default_values.StringType

        if N == None:  N = self.N
        if e == None:  e = self.e

        if not isinstance(M, StringType):  raise TypeError("'M' should be a string")

        M = [ord(char) for char in M]  # transfer into bytes

        # return [((value ** e) % N) for value in M]
        return [pow(value, e, N) for value in M]

    def decrypt(self, C, N=None, d=None):
        ListType = default_values.ListType

        if N == None:  N = self.N
        if d == None:  d = self.d

        if not isinstance(C, ListType):  raise TypeError("'C' should be a List")

        M = ""
        powered = 1

        for idx, value in enumerate(C):
            M += chr(pow(value, d, N))
            print(f"Iteration {idx + 1} \\ {len(C)}")

        return M

    @property
    def private_key(self):
        return {"d": self.d, "e": self.e}

    @property
    def public_key(self):
        return {"N": self.N, "e": self.e}

    @property
    def key(self):
        return {"N": self.N, "d": self.d, "e": self.e}

    def save_key_to_dir(self, key_to_save=None, dir=None, filename="key.rsa"):
        dictionaryType = default_values.DictionatyType
        if (not (key_to_save == None)):
            if not isinstance(key_to_save, dictionaryType):
                print("'key_to_save' needs to be a dictionary type")
        else:
            key_to_save = self.key

        dir += "\\" + str(filename)

        content = ""
        for i in key_to_save:
            content += "{} {} \n".format(i, self.key[i])

        file = open(dir, "w")
        file.write(content)
        file.close()

        print("saved")

    def save_public_key_to_dir(self, dir=None, filename="public_key.rsa"):

        dir += "\\" + str(filename)

        content = ""
        for i in self.public_key:
            content += "{} {} \n".format(i, self.key[i])

        file = open(dir, "w")
        file.write(content)
        file.close()

        print("saved")

    def save_private_key_to_dir(self, dir=None, filename="private_key.rsa"):

        dir += "\\" + str(filename)

        content = ""
        for i in self.private_key:
            content += "{} {} \n".format(i, self.key[i])

        file = open(dir, "w")
        file.write(content)
        file.close()

        print("saved")

    def load_keys(self, dir):
        with open(dir) as file:
            content = file.readlines()

        for i in content:
            if i[0] == "e":
                self.e = int(i[2:-2])
            if i[0] == "d":
                self.d = int(i[2:-2])
            if i[0] == "N":
                self.N = int(i[2:-2])

        print("loaded")

    def clear_keys(self):
        del self.e
        del self.d
        del self.N

    def print_public_key(self):
        print(f"N = {self.N}")
        print(f"e = {self.e}")

    def print_private_key(self):
        print(f"N = {self.N}")
        print(f"d = {self.d}")

    def print_key(self):
        print(f"N = {self.N}")
        print(f"e = {self.e}")
        print(f"d = {self.d}")


from numba import njit


# // faster RSA function to find the private key faster \\
@njit
def find_private_key(e, Phi):
    k = 1
    d = 5.5

    #while not (d.is_integer()):
    while not (int(d) == d):
        d = ((k*Phi) + 1) / e

        #if d >= Phi:
        #    raise ValueError("'d' cannot exceed phi({self.Phi})")

        if k % 1000000 == 0:
            print(k, d)

        k += 1

    print("final d =", int(d))
    return int(d)


import copy
import random
import numpy as np
import ML_functions
from pyalgo.basic_modules import default_functions, default_values
from pyalgo.basic_modules.pytrainer_default_values import ML_functions_trainer_default_values, ML_functions_default_values
from pyalgo import basic_modules


class Basic_ML_functions:

    """
    This class provides a set of basic functions to build your machine-learning model
    """

    def __init__(self): pass

    @staticmethod
    def Sigmoid(X):
        return 1/(1 + np.exp(-X))

    @staticmethod
    def LinearFunction(X):
        return X

    @classmethod
    def Tanh(cls, X, Scaler=2):
        floatType = default_values.FloatType
        intType = default_values.IntType
        if not ((isinstance(Scaler, floatType)) or (isinstance(Scaler, intType))):
            raise TypeError("'Scaler' needs to be a float")

        return 2 * cls.Sigmoid(Scaler * X) - 1

    @staticmethod
    def Relu(X):
        try:
            return max(0, X)
        except:
            m, n = X.shape
            return np.max(np.append(np.zeros((m, 1)), X, axis=1), axis=1).reshape((-1, 1))



class ML_functions_Training(Basic_ML_functions):

    def __init__(self): pass

    @staticmethod
    def Get_Randomized_Theta(epsilon, dim):
        FloatType = default_values.FloatType
        IntType = default_values.IntType


        if not (isinstance(epsilon, IntType) or isinstance(epsilon, FloatType)):
            raise TypeError("'epsilon' needs to be a int type or float type")

        randomized_theta = []

        Theta1d_size = 1

        for i in dim:
            Theta1d_size *= i

        for i in range(Theta1d_size):
            randomized_theta.append((random.random() * (2*epsilon)) - epsilon)


        randomized_theta = np.array(randomized_theta).reshape((dim))

        return randomized_theta

    @staticmethod
    def Normal_Equation(X, Y, AddBias=True):

        if AddBias:
            m, n = X.shape
            X = np.append(np.ones((m, 1)), X, axis=1)

        ArrayType = default_values.ArrayType

        if not isinstance(X, ArrayType):
            raise TypeError("X needs to be an array type")

        out = np.transpose(X).dot(X)
        out = np.linalg.inv(out)

        out = out.dot(np.transpose(X))

        out = out.dot(Y)

        return np.transpose(out)

    @staticmethod
    def forward_propagation(X, Theta, AddBias=True):

        self = ML_functions_Training

        X = copy.deepcopy(X)
        Theta = copy.deepcopy(Theta)

        num_of_matrices = 1  # // will be used for neural networks \\
        m, n = X.shape

        if AddBias:
            X = np.append(np.ones((m, 1)), X, axis=1)

        if len(Theta.shape) >= 3:
            print("needs to be completed")
            raise NotImplementedError("Yet...")  # // Needs to be built \\

        elif ((len(Theta.shape) == 1) or (len(Theta.shape) == 2)):
            print(Theta.shape)
            print("Simple Regression")
            print(f"num_of_matrices: {num_of_matrices}")
            return X.dot(Theta).reshape((-1, 1))

    @staticmethod
    def cost_function(X, Y, Theta, NonLinearityFunction=ML_functions.Basic_ML_functions.LinearFunction(), Method='MSE', AddBias=True, Lambda=None):

        ArrayType = default_values.ArrayType   # variables that helps check if the inputs are in the right type
        BoolType = default_values.BoolType     # variables that helps check if the inputs are in the right type

        # // check that the inputs are valid \\
        if not isinstance(X, ArrayType):
            raise TypeError("X needs to be an array type")

        if Method not in basic_modules.pytrainer_default_values.ML_functions_default_values.cost_function_Methods:
            raise ValueError(f"'Method' needs to be from the list of options: {ML_functions_default_values.cost_function_Methods}")

        if not isinstance(AddBias, BoolType):
            raise ValueError("The parameter 'AddBias' needs to be a boolean type")
        J = 0
        self = ML_functions_Training



        # // get the default value to lambda if the user didn't specified ant value to lambda \\
        if Lambda == None:
            Lambda = ML_functions_trainer_default_values.Lambda_default_value
        elif Lambda < 0:
            raise ValueError("Lambda can't be negative")


        m, n = X.shape

        J = NonLinearityFunction(self.forward_propagation(X, Theta, AddBias=AddBias) - Y)
        if Method == 'MAE':  J = abs(J)
        if Method == 'MSE':  J *= J

        J = np.sum(J) / m
        if Method == 'MSE':  J /= 2

        if Method == 'MAE':  J += (Lambda / m) * np.sum(Theta * Theta)
        if Method == 'MSE':  J += (Lambda / (m*2)) * np.sum(Theta * Theta)

        return J


from pyalgo.pytrainer.ML_functions import *

class trainer(ML_functions_Training):

    def __init__(self, add_bias=False, description=None):

        self.add_bias = add_bias
        self.description = description

    def print_model_settings(self):
        print(f'add_bias --> {self.add_bias}')
        print(f"description --> '{self.description}'")